/******************************************************************************/
/*  Detecting invalid privileges in RDB$USER_PRIVILEGES                       */
/*  Script invalidpriv.sql, created Dec 21, 2011 by Vitezslav Svejdar         */
/* Revised:                                                                   */
/*  Jul 19, 2013, cosmetic changes                                            */
/*   Jul 31, 2013 V 1.4:                                                      */
/*     Improved treatment of cases when SYSDBA or DBO grant something         */
/*     on an object that they do not own                                      */
/*   Aug 8, 2013 V 2.0:                                                       */
/*     No changes in this script                                              */
/******************************************************************************/
/*
This is an auxiliary script used in the process of substituting one db user
for another, described in the main script privprocs.sql.

The following table shows who can reassign ownership and who can grant or
revoke privileges to a table, a procedure, or a role.

        | Owner change                          | GRANT                         | REVOKE
--------------------------------------------------------------------------------------------------------
table   | Owner                                 | U with GO, SYSDBA             | U with GO, DBO, SYSDBA
proc    | Owner with Xpriv granted by himself   | Owner, SYSDBA                 | Owner, DBO, SYSDBA
role    | DBO, SYSDBA                           | Owner, U with AO, DBO, SYSDBA | Grantor, DBO, SYSDBA

Here Owner is the owner of the table (procedure, role) in question,
DBO is the database owner, GO is a shorthand for grant option,
AO is a shorthand for admin option. There is only one privilege on a procedure
(namely X, execute), there is only one privilege on a role (M, membership), but
there are several privileges on a table. Note that in particular
- neither SYSDBA nor the database owner can change ownership of a table
  or a stored procedure unless he is an owner of it,
- the ownership of a table, as well as the ownership of the entire database,
  is irrelevant for granting privileges on that table,
- privileges on a table are irrelevant for changing its ownership,
- on the other hand, the owner of a procedure or a role can grant a privilege on it,
- SYSDBA can grant and revoke any privilege,
- DBO can revoke any privilege.
Experiments show that there are some hidden rules about privileges:
(a) It can be a problem if some user A has a privilege on a table with grant option
granted by X, and the same privilege without grant option granted by Y.
(b) If a user is granted U or D on a table, Firebird expects that he is
also granted S.
These rules are not checked by the substitution process.
*/
select g.* from rdb$user_privileges g left join rdb$user_privileges x
on (x.rdb$user=g.rdb$grantor and x.rdb$user_type=8 -- Rights for a living person of the same name
and x.rdb$relation_name=g.rdb$relation_name and x.rdb$object_type=0 -- on the same table
and x.rdb$privilege=g.rdb$privilege and x.rdb$grant_option=1 -- the same right with GO
and (x.rdb$field_name is null or x.rdb$field_name=g.rdb$field_name) -- on all or the same field
) or g.rdb$grantor='SYSDBA' where g.rdb$object_type=0 -- Right on a TABLE or VIEW
and x.rdb$grantor is null -- Only some out of all invalid rights are essential:
and ((g.rdb$user='%0:s' and g.rdb$user_type=8) or g.rdb$grantor='%0:s')
union
select g.* from rdb$user_privileges g left join rdb$procedures p
on g.rdb$relation_name=p.rdb$procedure_name
where g.rdb$object_type=5 -- A right X on a procedure, i.e. the X privilege
and (p.rdb$owner_name is null -- non-existent procedure or
or (g.rdb$grantor<>p.rdb$owner_name and g.rdb$grantor<>'SYSDBA')) -- an illegal grantor
-- For us, a problem is an invalid right of the form A -> DUSER, A<>DUSER
and ((g.rdb$user='%0:s' and g.rdb$user_type=8 and g.rdb$grantor<>'%0:s')
-- but we do not mind rights DUSER -> B in case that B is an owner
or (g.rdb$grantor='%0:s' and p.rdb$owner_name<>'%1:s'))
union
select g.* from rdb$user_privileges g left join rdb$roles r on g.rdb$relation_name=r.rdb$role_name
left join rdb$user_privileges x on x.rdb$user=g.rdb$grantor and x.rdb$user_type=8
and x.rdb$relation_name=g.rdb$relation_name and x.rdb$object_type=13 and x.rdb$grant_option=2
left join rdb$relations t on t.rdb$relation_name='RDB$ROLES'
where g.rdb$object_type=13 -- A right on a role, i.e. the M privilege
and (r.rdb$owner_name is null -- non-existent role
or (x.rdb$grantor is null and g.rdb$grantor<>r.rdb$owner_name
and g.rdb$grantor<>'SYSDBA' and g.rdb$grantor<>t.rdb$owner_name))
-- ... (same as with procedures)
and ((g.rdb$user='%0:s' and g.rdb$user_type=8 and g.rdb$grantor<>'%0:s')
or (g.rdb$grantor='%0:s' and r.rdb$owner_name<>'%1:s'));
commit;
